function [data_cell param_cell code time_cell tpara_cell] = ... 
    retrieve(diag_name, shot_number, sub_shot, channels, varargin)
% Retrieve shot data with diagnostics,shot#,subshot# and channel#.
%   [data_cell param_cell code time_cell tpara_cell] = ... 
%      retrieve( diag_name, shot_number, sub_shot, channels, ... )
%
%       diag_name(text)      : name of diagnostics
%       shot_number(numeric) : SHOT number
%       sub_shot(numeric)    : SHOT sub-number
%       channels
%          (numeric [vector]): channel number(s)
%          If it wants to retrieve all channels, it is 0. 
%          If it wants to retrieve a channel, it is N. 
%          If it wants to retrieve a range, it is  START#:END#.
%                                              (e.g.  [1,3,5:7] )
%          (text)            : signal tag name (e.g. 'sig-tag1')
%
%       data_cell(cell)     : data for each channels
%       param_cell(cell)    : parameter for each channels
%       code(numeric)       : result code
%          If it is 0 , this function is success. otherwise, this
%          function is error.
%       time_cell(cell)     : time array for each channels
%       tpara_cell(cell)    : time's parameter for each channels
%
% Options :
%         'frame',frames(numeric [vector])
%                                : frame number
%             The format is the same as an channels parameter.
%         'host',hostname(text)  : retrieve from a direct server
%         'path',pathname(text)  : retrieve from a direct path
%                                  default: 'ShotDataFS'
%         'wait'                 : wait for the appearance of data
%             It can't be used with 'et','tt'.'ss'.
%         'timeout',timeout_sec(numeric)
%                                : seconds of timeout for wait
%             The default is 180 sec.
%         'v4'                   : Convert to voltage or real number
%                                  (single precision)
%         'vflt'                 : same 'v4'
%         'v8'                   : Convert to voltage or real number
%                                  (double precision)
%         'vdbl'                 : same 'v8'
%         'silent'               : do not display messages
%         'dataclass',classname(text)
%                                : fix the class of data.
%             'intN','uintN','single','double' (N=8,16,32,64)
%         'time-single'           : time data is a single type.
%
% Range Options :
%         'et',range(text)       : range by experiment timming
%             'N:M[u]' N:begin time, M:end time, u:unit(s,ms,us)(e.g. '1:2.5s' )
%         'tt',range(text)       : range by base of triggered time
%              format is same of 'et' option
%         'ss',range(text)       : range by sample number
%             'N:M' N:begin number, M:end number(e.g. '10000:20000' )
%
% Examples :
%    [data param code] = retrieve('SX8O',98000,1,0,'silent');
%    if code ~= 0
%       disp('Error');
%    else
%       printParam(param);
%    end
%
%
%    [data,~,code] = retrieve( 'SX8O', 98000,1,0,'wait','timeout',30 );
%    if code == 0
%       disp('Success');
%    end
%
%
%    >> [data,params,code,ta,tp] = retrieve('RADH',98000,1,[2:6,12],'v4','et','3:6s');
%    >> plot(ta{1},data{1});
%    >> printParam(params{1});
%
% Last update : 2020-06-10
%
% See also
%     printParam getParam retrieve_t changeSite currentSite retrieveVersion

%
% datasize = 0 : isempty(data{1}) is true.
%
  ip = inputParser;
  ip.KeepUnmatched = true;

  ip.addRequired('diag_name', @ischar);
  ip.addRequired('shot_number', @(x)isnumeric(x) && isscalar(x) && x > 0);
  ip.addRequired('sub_shot',@(x)isnumeric(x) && isscalar(x) && x >= 0);

% ip.addRequired('channels',@(x)isnumeric(x) && all(x>=0) );
% ip.addParamValue('frame',0,@(x)isnumeric(x) && x >= 0);

% ip.parse(diag_name, shot_number, sub_shot, channels);
  ip.parse(diag_name, shot_number, sub_shot);

  if nargout > 0
    data_cell = cell(0);
  end
  if nargout > 1
    param_cell = cell(0);
  end
  if nargout > 3
    time_cell = cell(0);
  end
  if nargout > 4
    tpara_cell = cell(0);
  end

  arg_check = true;
  verbose = true;
  wait_time_sec = 0;
  flg_timeout = false;
  opt_host = '';
  flg_host = false;
  opt_path = '';
  flg_path = false;
  opt_frame = 0;
  flg_frame = false;
  nvarargin = length(varargin);
  opt_vcnv = 0;       % 0=none 1=flt 2=dbl
  err_arg = '';
  opt_dataclass = '';
  flg_dataclass = false;
  is_time_double = 1;
  opt_range = '';
  opt_range_type = 0;
  flg_range = false;

  for idx = 1:nvarargin
    if isempty(varargin{idx})
      continue;
    elseif flg_timeout
      if isnumeric(varargin{idx}) && isscalar(varargin{idx})
        wait_time_sec = varargin{idx};
        flg_timeout = false;
        continue
      else
        arg_check = false;
        break
      end
    elseif flg_frame
      if isnumeric(varargin{idx}) && all(varargin{idx} >= 0)
        opt_frame = varargin{idx};
        flg_frame = false;
        continue
      else
        arg_check = false;
        break
      end
    elseif flg_dataclass
      if ischar(varargin{idx}) && any(strcmp(varargin{idx}, ...
              {'int8','int16','int32','int64',...
               'uint8','uint16','uint32','uint64','single','double'}))
        opt_dataclass = varargin{idx};
        flg_dataclass = false;
        continue
      else
        arg_check = false;
        break
      end
    elseif flg_range
      if 0 == length(opt_range)
        opt_range = varargin{idx};
        flg_range = false;
        continue
      else
        arg_check = false;
        break
      end
    end
    
    if ~ischar(varargin{idx}) 
      err_arg = strcat(err_arg,[' arg#' num2str(idx+4)]);
      arg_check = false;
      break
    end
    if flg_host
      opt_host = varargin{idx};
      flg_host = false;
      continue
    elseif flg_path
      opt_path = varargin{idx};
      flg_path = false;
      continue
    end
    switch varargin{idx}
      case 'frame'
        flg_frame = true;
      case 'host'
        flg_host = true;
      case 'path'
        flg_path = true;
      case 'timeout'
        flg_timeout = true;
      case 'silent'
        verbose = false;
      case 'wait'
        if wait_time_sec == 0
          wait_time_sec = 180;
        end
      case 'vflt'
        opt_vcnv = 1;
      case 'vdbl'
        opt_vcnv = 2;
      case 'v4'
        opt_vcnv = 1;
      case 'v8'
        opt_vcnv = 2;
      case 'dataclass'
        flg_dataclass = true;
      case 'time-single'
        is_time_double = 0;
      case 'et'
        flg_range = true;
        opt_range_type=1;
      case 'tt'
        flg_range = true;
        opt_range_type=2;
      case 'ss'
        flg_range = true;
        opt_range_type=3;
      otherwise
        arg_check = false;
        err_arg = strcat(err_arg, [' ''' varargin{idx} '''']);
    end
  end
  if not(ischar(channels) || ( isnumeric(channels) && all(channels>=0) ) )
    if verbose
      disp('<Error: illegal channles>');
      disp('Type "help retrieve".');
    end
    code = -2;
    return;
  end

  if( ~arg_check )
    if verbose
      disp(['<Error: illegal argument ' err_arg ' >']);
      disp('Type "help retrieve".');
    end
    code = -2;
    return;
  end
  if 0 < length(opt_host) && 0 == length(opt_path)
    opt_path='ShotDataFS';
  end
  [rtd, index_server] = mRetrieveOpen( diag_name, shot_number, sub_shot, ...
        opt_host, opt_path, wait_time_sec, opt_range_type, opt_range, verbose);
  if rtd < 0
    code=rtd;
    return;
  end
%
  [code last_channel_number syy smm sdd shour smin ssec trans_server] ...
      = mRetrieveShotInfo( rtd, verbose);
  if code < 0
    return
  end

  if ischar(channels)
    cnt_chs = 1;
  else
    if channels(1) == 0
      channels = 1:last_channel_number;
    end
    cnt_chs = length(channels);
  end
  if verbose 
    fprintf('---Arc Shot Infomation---\n');
    fprintf('Index Server Name  = %s\n' ,char(index_server)    );
    fprintf('Trans Server Name  = %s\n' ,char(trans_server)    );
    fprintf('Shot Number        = %d\n' ,shot_number           );
    fprintf('Sub Shot Number    = %d\n' ,sub_shot              );
    fprintf('Channel            = %d\n' ,last_channel_number    );
    fprintf('DATE               = %4d-%2d-%2d\n',syy, smm, sdd  );
    fprintf('TIME               = %4d:%2d:%2d\n',shour, smin, ssec );
    if ischar(channels)
      fprintf('Request Channel    = %s\n'    ,channels);
    else
      fprintf('Request Channel    = %3d:%3d\n'    ,min(channels), max(channels) );
    end
  end
  if not(ischar(channels)) && max(channels) > last_channel_number
    code = -2;
    if verbose
      fprintf('<Error: Illegal channel number!>\n');
    end
    mRetrieveClose(rtd);
    return
  end
  data_cell = cell(cnt_chs,1);
  if nargout > 1
    param_cell = cell(cnt_chs,1);
  end 
  data_idx = 0;
  for tmp_x = 1:cnt_chs
    data_idx = data_idx+1;
    if not(ischar(channels)) 
      ch_no = channels(data_idx); 
      [code data_bytes comp_bytes param_count data_type image_type value_len ...
            is_nframe ch_comment] = mRetrieveChInfo( rtd, ch_no, verbose);
    else
      [code data_bytes comp_bytes param_count data_type image_type value_len ...
            is_nframe ch_comment ch_no] = mRetrieveChInfo2( rtd, channels, verbose);
    end
    if code < 0
      return;
    end
% 
% if image_type is undefined, image_type remake with value_len
    if strcmp(image_type, '' )
      image_type = mValueLen2ImageType(value_len);
    end
    if verbose
      if 2 == data_type
        mes_type = 'ANA';
      else
        mes_type = 'RAW';
      end
      fprintf('---Channel Information---\n');
      fprintf('Channel Number     = %d\n'  ,ch_no         );
      fprintf('Data Type          = %s\n'  ,mes_type      );
      fprintf('Data Size          = %ld\n'  ,data_bytes    );
      fprintf('Comment            = %s\n'  ,ch_comment    );
      if 0 ~= comp_bytes
        fprintf('Compressed Size    = %ld\n'  ,comp_bytes);
      end
    end
    if ( is_nframe == 0 && opt_vcnv == 0 )
      [code data data_length] ...
         = mRetrieveChData( rtd, ch_no, data_bytes, image_type, opt_dataclass, verbose);
      if code < 0 
        break;
      else
        if data_length > 0 
          data_cell{data_idx} = data;
        end
      end
    elseif ( is_nframe == 0 && opt_vcnv ~= 0 )
      [code data data_length] ...
         = mRetrieveChV( rtd, ch_no, data_bytes, image_type, opt_vcnv, verbose);
      if code < 0 
          break;
      else
        if data_length > 0 
          data_cell{data_idx} = data;
        end
      end
    
    elseif ( is_nframe ~= 0 && opt_vcnv == 0 )
      if ~all( opt_frame <= is_nframe )
        code = -99;
        if verbose
          fprintf('<Error: not supported Volts convert with frame data!>\n');
        end
        break;
      end
      if opt_frame(1) == 0
        [code begin_no end_no] = mRetrieveRetrieveNo(rtd, ch_no);
        if( 0 == code )
          frame_nums = begin_no:end_no;
        else
          frame_nums = 1:is_nframe;
        end
      else
        frame_nums = opt_frame;
      end
      [code data data_length frame_x frame_y] ...
        = mRetrieveChFrameData( rtd, ch_no, frame_nums, verbose);
      if code == 1
        code = -99;
        if verbose
          fprintf('<Error: Each frame sizes differ. !>\n');
        end
        mRetrieveClose(rtd);
        return
      end
      if code < 0 
        break;
      else
        if data_length > 0
          f_num = length(frame_nums);
          data_cell{data_idx} = data;
%         if 1 == f_num && data_length ~= (frame_x*frame_y)
%            data_cell{data_idx} = data;
%         else
%            data_cell{data_idx} = reshape(data,frame_x,frame_y, f_num);
%         end
        end
      end
    else
      code = -99;
      if verbose
        fprintf('<Error: not supported Volts convert with frame data!>\n');
      end
      mRetrieveClose(rtd);
      return
    end
    if 1 > data_length
      if verbose
        fprintf('<Warning : CH#%d Data size is 0 !>\n',ch_no);
      end
    end
    if nargout < 2 % data only
      continue;
    end
    [code param_cell{data_idx}] = ...
        mRetrieveChParams( rtd, ch_no, param_count, verbose);
    if code < 0
      break;
    end
    if nargout < 4 
      continue;
    end
    [code time_dat data_length params] ...
      = mRetrieveChTime( rtd, diag_name, shot_number, sub_shot, ch_no, ...
        is_time_double, index_server, verbose);
    if code < 0 
     break;
    else
     if data_length > 0 
      time_cell{data_idx,1} = time_dat;
     end
    end
    if nargout < 5 
      continue;
    end
    tpara_cell{data_idx,1} = params;
  end

  mRetrieveClose(rtd, false);
  return
end

function [code data data_length frame_x frame_y] = ...
    mRetrieveChFrameData( rtd, ch_no, frame_nums, verbose)
% Get frame data in channel with opened retrieve-Protocol.
%   [code data_ptr data_length frame_x frame_y] = ...
%       mRetrieveChFrameData( rtd, ch_no, frame_nums, verbose)
%           farmes_nums   : '1:100', '2' not '0'
%

  data_length = 0;
  frame_num = length( frame_nums );
  data_cnt = 0;
  frame_x = 1;
  frame_y = 1;
  data = 0;
  for frame_no = frame_nums
    [code array_bytes image_type frame_x frame_y] = ...
        mRetrieveFrameInfo( rtd, ch_no, frame_no, verbose);
    if code < 0
      return;
    end
    [code fdata frame_length] = ...
        mRetrieveFrameData( rtd, ch_no, frame_no, frame_x, frame_y, image_type, verbose);
    if code < 0
      return;
    else
      if data_cnt == 0
        [data_class,byte_size] = mImageType2DataClass(image_type);
        if frame_num == 1 && frame_length ~= (frame_x*frame_y)
          data = zeros(1,frame_length, data_class);
        else
          data = zeros(frame_x,frame_y,frame_num, data_class);
        end
      end
%     start_p = data_cnt*(array_bytes/byte_size) + 1;
%     end_p = start_p + (array_bytes/byte_size) - 1;
      start_p = data_cnt*(frame_length) + 1;
      end_p = start_p + (frame_length) - 1;
      data(start_p:end_p) = fdata(1:frame_length);
    end
    
    if data_cnt == 0
      init_frame_length = frame_length;
    else
      if frame_length ~= init_frame_length
        code = 1;
        return
      end
    end
    data_length = data_length + frame_length;
    data_cnt = data_cnt+1;
  end
end

function image_type = mValueLen2ImageType( value_len )
% Make image_type value_len
%   image_type = mValueLen2ImageType( value_len )
%
  if value_len < 9
    image_type = 'INT8';
  elseif value_len < 17
    image_type = 'INT16';
  elseif value_len < 33
    image_type = 'INT32';
  else
    image_type = 'INT64';
  end
end

function [class_name bytes] = mImageType2DataClass( image_type )
% Make class name image_type
%   [class_name bytes] = mImageType2DataClass( image_type )
%
  switch image_type
    case 'INT8'
      bytes = 1;
      class_name='int8';
    case 'INT16'
      bytes = 2;
      class_name='int16';
    case 'INT32'
      bytes = 4;
      class_name='int32';
    case 'INT64'
      bytes = 8;
      class_name='int64';
    case 'FLT32'
      bytes = 4;
      class_name='single';
    case 'FLT64'
      bytes = 8;
      class_name='double';
    case 'YUV422'   % 2Byte
      bytes = 2;
      class_name='uint16';
    case 'YUY2'   % 2Byte 2020-06
      bytes = 2;
      class_name='uint16';
    case 'BayerBG8' % 1Byte 12bit
      bytes = 1;
      class_name='uint8';
    case 'BayerBG16'    % 2Byte
      bytes = 2;
      class_name='uint16';
    otherwise
      value_len = 0;
      if strncmp(image_type,'GRAY',4 ) || strncmp(image_type,'GREY',4 )
        value_len = str2double(image_type(5:length(image_type)));
      else
        if length(image_type) == 4 || length(image_type) == 3
          if strfind(image_type,'R') && strfind(image_type,'G') && ...
                  strfind(image_type,'B')
            value_len = 32;
          end
        end
      end
      if value_len < 9 
        bytes = 1;
        class_name = 'uint8';
      elseif value_len < 17 
        bytes = 2;
        class_name = 'uint16';
      elseif value_len < 33
        bytes = 4;
        class_name = 'uint32';
      else
        bytes = 8;
        class_name = 'int64';
      end
  end
end
function bytesize = mClassName2bytesize( class_name )
% 
%   bytesize = mClassName2bytesize( class_name )
%
  switch class_name
    case {'int8','uint8'}
      bytesize = 1;
    case {'int16','uint16'}
      bytesize = 2;
    case {'int32','uint32','single'}
      bytesize = 4;
    case {'int64','uint64','double'}
      bytesize = 8;
    otherwise
      bytesize = 1;
  end
end
function [code time_dat data_length params] = ...
    mRetrieveChTime( rtd, diag_name, shot_number, sub_shot, ch_no, is_double, index_server, verbose)
% Get Channel time data with opened retrieve-Protocol
%   [code time_dat data_length params] = ...
%       mRetrieveChTime( rtd, diag_name, shot_number, sub_shot, ch_no, is_double, index_server, verbose)
%

%
%RETRIEVE_API int retrieveTimeInfo( int rtr_d, int ch_no,
%					char* dts_source, char* dts_hostid, char*	dts_moduleid, char* dts_trig_ch,
%					char* clk_source, char* clk_hostid, char*	clk_moduleid, char* dts_clk_ch,
%					int* dts_user_define, char* clock_src, int64_t* time_array_size, 
%					void* clock_cycle, void* trigger_delay, void* sampling0timing, int is_double);
% [int32, cstring, cstring, cstring, cstring, cstring, cstring, cstring, cstring, int32Ptr, 
%  cstring, int64Ptr, voidPtr, voidPtr, voidPtr]  retrieveTimeInfo
% (int32, int32, cstring, cstring, cstring, cstring, cstring, cstring, cstring, cstring, 
%  int32Ptr, cstring, int64Ptr, voidPtr, voidPtr, voidPtr, int32)
%
%RETRIEVE_API int retrieveTimeData( int rtr_d, int ch_no,
%			void* data, int64_t array_size, int64_t* data_length,int is_double );
%[int32, voidPtr, int64Ptr]	retrieveTimeData	(int32, int32, voidPtr, int64, int64Ptr, int32)
%
  code = 0;
  time_dat = 0;
  data_length = 0;
  if nargout > 3
   params(13,1) = struct('Pname',[],'Pvalue',[],'Ptype',[]);
  end

  if verbose
    fprintf('---DTS Infomation---\n');
  end
  for n=1
    [code dts_source dts_host	dts_mod dts_trg_ch ...
         clk_source clk_host	clk_mod dts_clk_ch ...
         dts_user_define ext_or_int	time_array_size ...
         clock_cycle trig_delay trig_timing ] ...
      = cRetrieveTimeInfo( rtd, ch_no, is_double, verbose);
    if code ~= 0
      break;
    end
    if verbose
      fprintf('DTS source         = %s\n', dts_source);
      fprintf('DTS host ID        = %s\n', dts_host);
      fprintf('DTS module ID      = %s\n', dts_mod);
      fprintf('DTS trigger ch     = %s\n', dts_trg_ch);
      fprintf('CLK source         = %s\n', clk_source);
      fprintf('CLK host ID        = %s\n', clk_host);
      fprintf('CLK module ID      = %s\n', clk_mod);
      fprintf('DTS clock ch       = %s\n', dts_clk_ch);
      fprintf('Clock Source       = %s\n', ext_or_int);
      fprintf('DTS user define    = %10ld\n', dts_user_define);
      fprintf('Time array size    = %10ld\n', time_array_size);
    end

    [code time_dat data_length] = ...
        cRetrieveTimeData( rtd, ch_no, time_array_size, is_double, verbose);
    if code ~= 0 
      break;
    end
    
    if verbose 
      fprintf('Cycle   = [%f] sec\n',clock_cycle);
      fprintf('Delay   = [%f] sec\n',trig_timing);
    end
    
    if nargout > 3
%     params(13,1) = struct('Pname',[],'Pvalue',[],'Ptype',[]);

      params(1,1) = struct('Pname', 'DTSsource',    'Pvalue', dts_source, 'Ptype', 1);
      params(2,1) = struct('Pname', 'DTShostID',    'Pvalue', dts_host, 'Ptype', 1);
      params(3,1) = struct('Pname', 'DTSmoduleID',  'Pvalue', dts_mod, 'Ptype', 1);
      params(4,1) = struct('Pname', 'TriggerCh',    'Pvalue', dts_trg_ch, 'Ptype', 4);
      params(5,1) = struct('Pname', 'CLKsource',    'Pvalue', clk_source, 'Ptype', 1);
      params(6,1) = struct('Pname', 'CLKhostID',    'Pvalue', clk_host, 'Ptype', 1);
      params(7,1) = struct('Pname', 'CLKmoduleID',  'Pvalue', clk_mod, 'Ptype', 1);
      params(8,1) = struct('Pname', 'ClockCh',      'Pvalue', dts_clk_ch, 'Ptype', 4);
      params(9,1) = struct('Pname', 'ClockSource',  'Pvalue', ext_or_int, 'Ptype', 1);
      params(10,1) = struct('Pname', 'TimeArraySize','Pvalue', num2str(time_array_size), 'Ptype', 7);

      if is_double == 1
        params(11,1) = struct('Pname', 'ArrayDataType',   'Pvalue', 'Double', 'Ptype', 1);
        params(12,1) = struct('Pname', 'ClockCycle(sec)', 'Pvalue', num2str(clock_cycle, 15), 'Ptype', 6);
        params(13,1)= struct('Pname', 'StartTiming(sec)','Pvalue', num2str(trig_timing, 15), 'Ptype', 6);
      else
        params(11,1) = struct('Pname', 'ArrayDataType',   'Pvalue', 'Single', 'Ptype', 1);
        params(12,1) = struct('Pname', 'ClockCycle(sec)', 'Pvalue', num2str(clock_cycle, 7), 'Ptype', 5);
        params(13,1)= struct('Pname', 'StartTiming(sec)','Pvalue', num2str(trig_timing, 7), 'Ptype', 5);
      end
    end
    break;
  end % end loop
  if code ~= 0 
    mRetrieveClose( rtd, false);
  end
end
