#include <octave/oct.h>
#include <octave/oct-map.h>
#include <octave/uint32NDArray.h>
#include "dbstore-oct.h"

/**
DBSTORE_API int dbsCloseFrame4(int dbs_d, int ch_no, 
				   unsigned int param_cnt, char* param_item[], char* param_val[], int param_type[])
 */

DEFUN_DLD (dbsCloseFrame, args, nargout,
" Close Process for write frame data in channel. \n"
"   code = dbsCloseFrame( dbs_des, channel_number, param_struct, ...) \n"
" \n"
"       dbs_des(numeric)    : opened dbStore's descriptor id \n"
"       channel_number(numeric) \n"
"                           : channel number \n"
"       param_struct(struct): parameter of channel \n"
"           struct fields are Pname(char),Pvalue(char),Ptype(numeric). \n"
"           describe it with 'struct([])' when there are not parameter to \n"
"           register. \n"
" \n"
"       code(numeric)       : status, 0 is success, other is error. \n"
" \n"
"  OPTIONS : \n"
"       'verbose'           : If an error occurs, output messages to display. \n"
" \n"
"  Example : \n"
"       des = dbsOpen( .... ); \n"
"       if des > 0 \n"
"          while true \n"
"              ch=1; \n"
"              code = dbsOpenFrame(des, ch, 640,480,'RGBA'); \n"
"              if code ~= 0  \n"
"                 break; \n"
"              end \n"
"              code = dbsWriteFrame(des, ch, 1, frame_array);  \n"
"              if code ~= 0  \n"
"                 break; \n"
"              end \n"
"  \n"
"              param=initParanList(); \n"
"              .... \n"
"              code = dbsCloseFrame(des, ch, param);  \n"
"              if code ~= 0  \n"
"                 break; \n"
"              end \n"
"              .... \n"
"          end \n"
"          .... \n"
"          dbsClsoe(des); \n"
"       end \n"
" \n"
"  Note : \n"
"    IF parameter is nothing, param_struct = struct([]). \n"
" \n"
"  Last update : 2018-05-30 \n"
" \n"
"  See also \n"
"      dbsOpen(.oct) dbsWrite(.oct) dbsClose(.oct) dbsAbort(.oct) \n"
"      initParamList(.m) setParam(.m) printParam(.m) \n"
"      dbsOpenFrame(.oct) dbsWriteFrame(.oct) dbsCloseFrame(.oct) \n" 
"      dbsWriteFrames(.oct) dbsMessage(.oct) dbstoreVersion(.oct) dbstore.txt \n"
           )
{
  const int MIN_IN_ARG = 3;
  const int NUM_OUT_ARG = 1;
  const std::string funcn = "dbsCloseFrame";

  int dbsd;
  int ch_no;
  octave_map params_map;
  PARAMS_BASE params_p;
  
  std::string opt_key = "";
  bool opt_verbose = false;
  
  int nargin = args.length();
  int code = 0;
  params_base_initial( params_p);
  
  for( ;; ) {
    // check 0.nargin
    if( nargin < MIN_IN_ARG ) { code = OCT_IF_ERR_ARG_NUM_INPUT_ARG;  break; }
    // check 0. nargout
    if( nargout > NUM_OUT_ARG ) { code = OCT_IF_ERR_ARG_NUM_OUTPUT_ARG;  break; }
    
    // check 1.dbsd
    if( 0 != (code = get_arg_int( args(0), dbsd )) )  {
      code += OCT_IF_ERR_ARG1;  break;
    }
    if( retrieve_debug ) octave_stdout << funcn <<" dbs_des:" << dbsd << "\n";
    // check 2.channel_number
    if( 0 != (code = get_arg_int( args(1), ch_no )) )  {
      code += OCT_IF_ERR_ARG2;  break;
    }
    if( retrieve_debug ) octave_stdout << funcn <<" channel_number:" << ch_no << "\n";
    // check 3.param_struct
#if (OCTAVE_MAJOR_VERSION == 4) && (OCTAVE_MINOR_VERSION == 2)
    if ( !args(2).is_map() )  {
#else
    if ( !args(2).isstruct() ) {
#endif
      code = OCT_IF_ERR_ARG3 + OCT_IF_ERR_PARAMETER_TYPE_MISSMATCH_STRUCT;   break;
    }
    //    scalar_map_value s
    params_map = args(2).map_value();
    if( 0 != (code=params_from_octave( params_map, params_p)) ) {
      code += OCT_IF_ERR_ARG3; break;
    }
    break;
  } // End of for( ;; )
  // Option
  int xcode = 0;
  for( int x=MIN_IN_ARG; x < nargin; x++ ) {
    if( 0 == (xcode=get_arg_string( args(x), opt_key )) ) {
      if( opt_key == "verbose" ) {
        opt_verbose = true;
      }
      else {
        if( 0 == code )
          code = OCT_IF_ERR_ARG_OPT(x+1) + OCT_IF_ERR_ARG_NO_SUPPORTED_OPTION;
      }
    }
    else {
      if( 0 == code )
        code = OCT_IF_ERR_ARG_OPT(x+1) + xcode;
    }
  } // End of Option

  if( 0 == code ) {
    code = dbsCloseFrame4( dbsd, ch_no, params_p.paramCounts, 
                    params_p.paramItem, params_p.paramVal, params_p.paramType);
  }
  params_base_free( params_p );
  
  octave_value_list ret_vals;
  ret_vals(0) = code;
  if( 0 != code ) {
    if( opt_verbose || retrieve_debug ) octave_dbs_error_message(code, funcn);
  }
  return ret_vals;
}
